import subprocess


def run_evolver(se_input_filepath: str, evolver_filepath: str = "evolver ") -> int:
    """
    Run the Surface Evolver file using the Surface Evolver software

    :param se_input_filepath: Path to Surface Evolver file
    :rtype se_input_filepath: str
    :param evolver_filepath: Path to the Surface Evolver interpreter
    :rtype evolver_filepath: str
    :return: Return code of the subprocess.run() function
    :rtype: int
    """
    p = subprocess.run([evolver_filepath, se_input_filepath], stdout=subprocess.DEVNULL)
    return p.returncode == 0



from dataclasses import dataclass
import numpy as np
import scipy.spatial as spatial
from copy import deepcopy


@dataclass
class Lattice:
    """
    Lattice class with all lattice properties

    :param number_cells_x: Number of cells in the x-axis to create
    :type  number_cells_x: int
    :param number_cells_y: Number of cells in the x-axis to create
    :type  number_cells_y: int
    """
    number_cells_x: int
    number_cells_y: int

    tessellation: object = None

    def __post_init__(self):
        ...

    def generate_square_seeds(self, standard_deviation: float = 0, spatial_step: int = 1) -> list:
        """
        Generate seeds for tessellation as a rectangular grid

        :param standard_deviation: Move seeds with a normal with this standard deviation
        :type standard_deviation: float
        :param spatial_step:
        :type spatial_step: int
        :return: List of [x,y] coordinates for the seeds
        :rtype: list
        """
        grid_values = ([[(i + np.random.normal(0, standard_deviation)) * spatial_step,
                         (j + np.random.normal(0, standard_deviation)) * spatial_step]
                        for j in range(0, self.number_cells_y) for i in range(0, self.number_cells_x)])
        return grid_values

    def generate_voronoi_tessellation(self, seed_values: list) -> object:
        """
        Generates the voronoi tessellation from the seed values provided

        :param seed_values: [x, y] list of positions to use as seeds for the voronoi tessellation
        :type seed_values: list
        :return: Voronoi tessellation object generated from the seed values
        :rtype: scipy.spatial.Voronoi()
        """
        self.tessellation = spatial.Voronoi(list(seed_values))

        return self.tessellation

    def create_lattice_elements(self) -> tuple:
        """
        Create the vertices, edges and cells from the scipy.spatial.Voronoi() objects

        :return: Vertices, edges and cell's list
        :rtype: tuple
        """
        new_vertices = {}
        new_cells = {}
        new_edges = {}
        regions = deepcopy(self.tessellation.regions)
        big_edge = []

        cnum = 1

        regions = self.remove_infinite_regions(regions)

        for c in regions:
            temp_for_cell = []
            temp_vertex_for_cell = []
            if len(c) != 0 and -1 not in c:
                # add first to close the cell_vertices
                c.append(c[0])
                for ii in range(0, len(c) - 1):
                    temp_big_edge = []
                    x_coordinate = np.around(np.linspace(round(self.tessellation.vertices[c[ii]][0], 3),
                                                         round(self.tessellation.vertices[c[ii + 1]][0], 3), 2), 3)
                    y_coordinate = np.around(self.line_eq(self.tessellation.vertices[c[ii]],
                                                          self.tessellation.vertices[c[ii + 1]],
                                                          x_coordinate), 3)

                    new_edge_vertices = list(zip(x_coordinate, y_coordinate))
                    # add new edges to the global list
                    for v in range(0, len(new_edge_vertices) - 1):
                        v0 = new_edge_vertices[v]
                        v1 = new_edge_vertices[v + 1]

                        vertex_number_1 = self.get_vertex_number(v0, new_vertices)

                        vertex_number_2 = self.get_vertex_number(v1, new_vertices)

                        enum = self.get_enum([vertex_number_1, vertex_number_2], new_edges)

                        temp_big_edge.append(enum)
                        temp_for_cell.append(enum)
                        temp_vertex_for_cell.append(vertex_number_1)
                        temp_vertex_for_cell.append(vertex_number_2)

                    big_edge.append(temp_big_edge)

                area_sign = self.get_cell_area_sign(temp_vertex_for_cell, new_vertices)
                new_cells[-1 * cnum * area_sign] = temp_for_cell
                # new_cells[cnum] = temp_for_cell
                # calculate cell_vertices centroid
                cnum += 1

        return new_vertices, new_edges, new_cells

    def get_middle_cells(self):
        # TODO implement function
        return None

        # middle_cells = []
        # other_cells = []
        # for c in cells:
        #     cnum = cells.index(c)
        #     xcm = np.mean([self.tessellation.vertices[ii][0] for ii in c])
        #     ycm = np.mean([self.tessellation.vertices[ii][1] for ii in c])
        #     # print(xcm)
        #     # if xcm > 80:
        #     #     right_cells.append(-1 * cnum * area_sign)
        #     # else:
        #     #     left_cells.append(-1 * cnum * area_sign)
        #     # take the middle cells
        #     if xcm > 60 and xcm < 110:
        #         middle_cells.append(cnum)
        #     else:
        #         other_cells.append(cnum)
        #
        # return middle_cells, other_cells

    def remove_infinite_regions(self, regions: list, max_distance: float = 50) -> list:
        """
        Remove regions that have a vertex too far away to be part of the tissue. This solves vertices placed in
        'infinity' by the tessellation

        :param regions: List of the voronoi regions
        :type regions: list
        :param max_distance: Maximum distance of a vertex in a polygon
        :type max_distance: float, optional
        :return: List of regions without the offending ones
        :rtype: list
        """
        to_delete = []
        for c in regions:
            distances = []
            if len(c) != 0 and -1 not in c:
                for ii in range(0, len(c) - 1):
                    distances.append(
                        np.linalg.norm(self.tessellation.vertices[c[ii]] - self.tessellation.vertices[c[ii + 1]]))
                distances = np.array(distances)
                if np.any(np.where(distances > max_distance, True, False)):
                    to_delete.append(c)

        for c in to_delete:
            regions.remove(c)
        return regions

    def get_cell_area_sign(self, cell: list, all_vertices: dict) -> int:
        """
        Get the orientation of the polygon for a cell_vertices using the sign of the area through gauss formula

        :param cell: Vertices of the cell_vertices to determine orientation
        :type cell: list
        :param all_vertices: Dictionary with all the vertices in the lattice
        :type all_vertices: dict
        :return: Sign of the cell_vertices area to determine orientation of the polygon
        :rtype: int
        """
        return int(np.sign(self.get_cell_area(cell, all_vertices)))

    def create_example_lattice(self, voronoi_seeds_std: float = 0.15, voronoi_seeds_step: int = 20) -> tuple[dict, dict, dict]:
        """
        Create a lattice with the initial Voronoi tessellation

        :param voronoi_seeds_std: Noise taken from a normal of mean zero and this value as standard deviation
        :type voronoi_seeds_std: float
        :param voronoi_seeds_step: SSpatial step to deposition of seeds in the tessellation
        :type voronoi_seeds_step: float
        :return: The vertices, edges and cells generated
        :rtype: tuple
        """
        self.generate_voronoi_tessellation(
            self.generate_square_seeds(standard_deviation=voronoi_seeds_std,
                                       spatial_step=voronoi_seeds_step))
        vertices, edges, cells = self.create_lattice_elements()

        return vertices, edges, cells

    @staticmethod
    def get_coordinates(vertices: dict) -> tuple:
        """
        Get the coordinates for all vertices in the system

        :param vertices: List of vertices
        :type vertices: dict
        :return: X and Y coordinates
        :rtype: tuple
        """
        x_coordinates, y_coordinates = zip(*vertices.values())

        return list(x_coordinates), list(y_coordinates)

    @staticmethod
    def get_edge_centroid(edge_id: int, vertices: list, edges: list) -> tuple:
        """
        Get the center of the requested edge

        :param edge_id: id of edge
        :type edge_id: int
        :param vertices: List of vertices
        :type vertices: list
        :param edges: list of edges
        :type edges: list
        :return: x and y coordinates of the edge's centroid
        """
        x_coordinates = [vertices[ii][0] for ii in edges[edge_id]]
        y_coordinates = [vertices[ii][1] for ii in edges[edge_id]]
        xm = np.mean(x_coordinates)
        ym = np.mean(y_coordinates)
        return xm, ym

    @staticmethod
    def get_normally_distributed_densities(edges: dict, center: float = 1, standard_deviation: float = 0.01) -> dict:
        """
        Get a dictionary of tensions that is normally distributed around the center with a certain standard deviation

        :param edges: List of edges in the lattice
        :type edges: dict
        :param center: Center of the distribution of tensions
        :type center: float
        :param standard_deviation: Standard deviation for the distribution of tensions
        :type standard_deviation: float
        :return: Dictionary with the keys as edge ids and the tensions as values
        :rtype: dict
        """
        return {k: np.random.normal(center, standard_deviation) for k, _ in edges.items()}

    @staticmethod
    def get_normally_distributed_volumes(cells: dict, means=(500,), stds=(50,), weights=None) -> dict:
        assert len(means) == len(stds)
        if len(means) > 1:
            assert (weights is None) or (
                        len(weights) == len(means)), f'"weights" and "values" have to be of equal length'
            v = np.random.choice(len(means), len(cells), p=weights)
            return {k: int(np.random.normal(means[v[i]], stds[v[i]])) for i, k in enumerate(cells.keys())}
        return {k: int(np.random.normal(means[0], stds[0])) for k in cells.keys()}

    @staticmethod
    def get_vertex_number(vertex: list, vertices: dict) -> int:
        """
        Get id of the vertex from the vertex position or new possible ID if the vertex is already known

        :param vertex: Vertex position
        :type vertex: list
        :param vertices: Dictionary of the vertices currently identified
        :type vertices: dict
        :return: Vertex id if the vertex exists, if not the next possible id to assign
        :rtype: int
        """
        if vertex in vertices.values():
            vertex_number = list(vertices.keys())[list(vertices.values()).index(vertex)]
        else:
            if len(vertices) > 0:
                vertex_number = max(vertices.keys()) + 1
            else:
                vertex_number = 1
            vertices[vertex_number] = vertex
        return vertex_number

    @staticmethod
    def get_enum(edge: list, edges: dict) -> int:
        """
        Get id of the edge from the vertex position or new possible ID if the vertex is already known

        :param edge: id of the vertices that create the edge
        :type edge: list
        :param edges: Dictionary of the edges currently identified
        :type edges: dict
        :return: edge id if the edge exists, if not the next possible id to assign
        :rtype: int
        """
        if edge in edges.values():
            enum = list(edges.keys())[list(edges.values()).index(edge)]
        elif edge[::-1] in edges.values():
            enum = - list(edges.keys())[list(edges.values()).index(edge[::-1])]
        else:
            if len(edges) > 0:
                enum = max(edges.keys()) + 1
            else:
                enum = 1
            edges[enum] = [edge[0], edge[1]]
        return enum

    @staticmethod
    def line_eq(p0: float, p1: float, x: np.ndarray) -> list:
        """
        Get the value of linear function that goes through p0 and p1 at x

        :param p0: First point to join
        :type p0: float
        :param p1: Second point to join
        :type p0: float
        :param x: Independent variable of the equation
        :type p0: list
        :return: Value of y(x) for a line that goes through p0 and p1
        """
        p0 = np.around(p0, 3)
        p1 = np.around(p1, 3)
        m = (p1[1] - p0[1]) / (p1[0] - p0[0])
        return p0[1] + m * (x - p0[0])

    @staticmethod
    def get_cell_area(cell_vertices: list, vertices: dict) -> float:
        """
        Area of a polygon using the shoelace formula

        :param cell_vertices: List of vertex's id of the polygon
        :param vertices: Dictionary of vertices in the system
        :return: Area of the polygon. Positive number indicates clockwise orientation, negative number
        counterclockwise.
        """
        x = [vertices[i][0] for i in cell_vertices]
        y = [vertices[i][1] for i in cell_vertices]
        return 0.5 * (np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))



import numpy as np
import scipy.stats as sps
import seapipy.lattice_class as lattice_class
import seapipy.surface_evolver as surface_evolver


class ExampleTissues:
    """
    Parent class for creating tissues

    :param parameters: Dictionary with all the relevant parameters of the system
    :type parameters: dict
    """
    def __init__(self, parameters):
        self.cells = None
        self.edges = None
        self.vertices = None

        self.parameters = parameters

        self.lattice = self.create_lattice()

        self.volumes, self.densities = self.get_initial_densities()
        self.se_object = self.create_se_file()

        self.se_object.generate_fe_file()

        self.evolve()

    def create_lattice(self):
        """
        Create a lattice with the initial Voronoi tessellation

        :return: Lattice class object with the initial Voronoi tessellation
        :rtype: lattice_class.Lattice
        """
        # TODO Use the lattice_class.create_example_lattice() instead of this function
        lattice = lattice_class.Lattice(self.parameters["n_cells_x"], self.parameters["n_cells_y"])

        lattice.generate_voronoi_tessellation(
            lattice.generate_square_seeds(
                standard_deviation=self.parameters["voronoi_seeds_std"],
                spatial_step=self.parameters["voronoi_seeds_step"]
            )
        )
        self.vertices, self.edges, self.cells = lattice.create_lattice_elements()
        return lattice

    def create_se_file(self):
        """
        Create the initial Surface Evolver clean slate to start writing to it

        :return: Surface Evolver object that will be written to disk
        :rtype: surface_evolver.SurfaceEvolver
        """
        se_object = surface_evolver.SurfaceEvolver(self.vertices,
                                                        self.edges,
                                                        self.cells,
                                                        self.densities,
                                                        self.volumes,
                                                        polygonal=False)

        return se_object

    def get_initial_densities(self):
        """
        Create dictionary of cell volumes and membrane densities to be used in the initial Surface Evolver simulation
        using the parameters provided to the class

        :return: cell volumes and membrane densities
        """
        cell_volumes = self.lattice.get_normally_distributed_volumes(self.cells,
                                                                     means=(self.parameters['cell_v_mean'],),
                                                                     stds=(self.parameters['cell_v_std'],),
                                                                     weights=None)

        densities = self.lattice.get_normally_distributed_densities(self.edges,
                                                                    1,
                                                                    0.1)

        return cell_volumes, densities

    def evolve(self):
        """
        Add the initial evolution steps to the Surface Evolver slate

        :return: None
        """
        self.se_object.initial_relaxing(evolve_step=10000)
        self.se_object.evolve_relaxing(10, 2500)
        self.se_object.add_vertex_averaging(100)
        self.se_object.change_scale(0.005)
        self.se_object.evolve_relaxing(5, 5000)
        self.se_object.save_one_step(self.parameters['save_dir'], self.parameters['file_name'])
        self.se_object.evolve_relaxing(5, 5000)

    def save_many_steps(self, max_steps: int = 50, step: int = 50):
        """
        Add to the Surface Evolver slate saving loop *max_steps* times every *step* steps

        :param max_steps: Number of saves to generate
        :type max_steps: int
        :param step: Evolution steps between savings
        :type step: int
        :return: None
        """
        self.se_object.save_many_steps(self.parameters['save_dir'], self.parameters['file_name'], max_steps, step)


class NormalFurrow(ExampleTissues):
    """
    Child class of :class:`ExampleTissues` to create a tissue with a normal distribution of membrane densities in the
    horizontal or vertical axis

    :param parameters: Dictionary with all the relevant parameters of the system
    :type parameters: dict
    """
    def __init__(self, parameters):
        super().__init__(parameters)

        self.new_densities = self.get_new_densities(axis=self.parameters["axis"])
        self.se_object.change_line_tensions(self.new_densities)
        self.se_object.save_one_step(self.parameters['save_dir'], self.parameters['file_name'])

    def get_new_densities(self, axis="x"):
        """
        Create the densities dictionary with the normal distribution in the horizontal or vertical axis

        :param axis: x or y
        :type axis: str
        :return: Dictionary with the new densities to be assigned to the membranes
        :rtype: dict
        """
        tissue_center = np.mean(self.lattice.get_coordinates(self.vertices), axis=1)
        tissue_min = np.min(self.lattice.get_coordinates(self.vertices), axis=1)
        tissue_max = np.max(self.lattice.get_coordinates(self.vertices), axis=1)
        if axis == "x":
            xrange = np.arange(tissue_min[0], tissue_max[0])
            distribution_peak = np.max(sps.norm.pdf(xrange, loc=tissue_center[0], scale=self.parameters['edge_t_std']))
            norm_value = {k: sps.norm.pdf(self.lattice.get_edge_centroid(k, self.vertices, self.edges)[0],
                                          tissue_center[0],
                                          self.parameters['edge_t_std']) / distribution_peak
                          for k, _ in self.edges.items()}

            new_densities = {k: self.densities[k] + self.parameters['edge_t_mean'] * norm_value[k]
                             for k, edge in self.edges.items()}
        elif axis == "y":
            xrange = np.arange(tissue_min[1], tissue_max[1])
            distribution_peak = np.max(sps.norm.pdf(xrange, loc=tissue_center[1], scale=self.parameters['edge_t_std']))
            norm_value = {k: sps.norm.pdf(self.lattice.get_edge_centroid(k, self.vertices, self.edges)[1],
                                          tissue_center[1],
                                          self.parameters['edge_t_std']) / distribution_peak
                          for k, _ in self.edges.items()}
            new_densities = {k: self.densities[k] + self.parameters['edge_t_mean'] * norm_value[k]
                             for k, edge in self.edges.items()}
        else:
            raise NotImplementedError

        return new_densities


class CircularFurrow(ExampleTissues):
    """
    Child class of :class:`ExampleTissues` to create a tissue with a normal distribution of membrane densities at a fixed distance from the center
    of the tissue. Effectively, gives a circular furrow.

    :param parameters: Dictionary with all the relevant parameters of the system
    :type parameters: dict
    """
    def __init__(self, parameters):
        super().__init__(parameters)

        self.volumes, self.densities = self.get_initial_densities()
        self.se_object = self.create_se_file()

        self.se_object.generate_fe_file()
        self.evolve()
        self.new_densities = self.get_new_densities()
        self.se_object.change_line_tensions(self.new_densities)
        self.se_object.save_one_step(self.parameters['save_dir'], self.parameters['file_name'])

    def get_new_densities(self):
        """
        Create the densities dictionary with the normal distribution in a circular furrow

        :return: Dictionary with the new densities to be assigned to the membranes
        :rtype: dict
        """
        new_densities = {}
        tissue_center = np.mean(self.lattice.get_coordinates(self.vertices), axis=1)

        for k, edge in self.edges.items():
            centroid = self.lattice.get_edge_centroid(k, self.vertices, self.edges)
            distance = np.linalg.norm(tissue_center - centroid)
            distribution_peak = sps.norm.pdf(0, loc=0, scale=self.parameters['edge_t_std'])

            new_densities[k] = sps.norm.pdf(distance,
                                            loc=0,
                                            scale=self.parameters['edge_t_std']) / distribution_peak
        # make densities have an average of 1:
        mean_value = np.mean(list(new_densities.values()))
        new_densities = {k: value / mean_value for k, value in new_densities.items()}
        return new_densities


class RandomCellTypes(ExampleTissues):
    """
    Child class to create a tissue with a random assignment of density to each membrane

    :param parameters: Dictionary with all the relevant parameters of the system
    :type parameters: dict
    """
    def __init__(self, parameters):
        super().__init__(parameters)

        self.volumes, self.densities = self.get_initial_densities()
        self.se_object = self.create_se_file()

        self.se_object.generate_fe_file()
        self.evolve()
        self.new_densities = self.get_new_densities()

        self.se_object.change_line_tensions(self.new_densities)
        self.se_object.save_one_step(self.parameters['save_dir'], self.parameters['file_name'])

    def get_new_densities(self):
        """
        Create the densities dictionary using a random assignment of values. It can be controlled using the parameters
        dictionary.

        :return: Dictionary with the new densities to be assigned to the membranes
        :rtype: dict
        """
        new_densities = {}
        for k, _ in self.edges.items():
            chosen = np.random.choice(self.parameters["edge_tensions"])
            scale = self.parameters["edge_tensions_std"] * chosen

            new_densities[k] = np.random.normal(loc=chosen, scale=scale)

        return new_densities



from dataclasses import dataclass
import io


@dataclass
class SurfaceEvolver:
    """
    Class to interact with the Surface Evolver file buffer.

    :param vertices: Vertices in the tesselation
    :type vertices: dict
    :param edges: Edges in the tesselation
    :type edges: dict
    :param vertices: Cells in the tesselation
    :type cells: dict
    :param density_values: Initial line tensions for the edges
    :type density_values: dict
    :param volume_values: Initial target areas in the tesselation
    :type volume_values: dict
    :param polygonal: Whether to use polygons or allowed curved edges
    :type polygonal: bool, optional
    """
    vertices: dict
    edges: dict
    cells: dict

    density_values: dict
    volume_values: dict

    polygonal: bool = True

    def __post_init__(self):
        self.fe_file = io.StringIO()
        self.density_values = {key: round(value, 3) for key, value in self.density_values.items()}

    def generate_fe_file(self) -> io.StringIO:
        """
        Generate the initial Surface Evolver slate to write into

        :return: Initialized Surface Evolver slate
        :rtype: io.StringIO()
        """
        self.fe_file.write("SPACE_DIMENSION 2 \n")
        self.fe_file.write("SCALE 0.005 FIXED\n")
        self.fe_file.write("STRING \n")
        self.fe_file.write("\n")
        self.fe_file.write("vertices \n")

        for k, v in self.vertices.items():
            self.fe_file.write(f"{k}   {round(v[0], 3)} {round(v[1], 3)}\n")
        self.fe_file.write("\n")
        self.fe_file.write("edges \n")
        for k, v in self.edges.items():
            # lambda_val = round(new_edges_tensions[[k in ee for ee in bedge].index(True)], 3)
            lambda_val = self.density_values[k]
            self.fe_file.write(f"{abs(k)}   {v[0]}   {v[1]}   density {lambda_val}\n")

        self.fe_file.write("\n")
        self.fe_file.write("faces \n")
        for k, v in self.cells.items():
            # color = np.random.choice(list(color_list))
            str_value = " ".join(str(vv) for vv in v)
            self.fe_file.write(f"{abs(k)}   {str_value} \n")

        self.fe_file.write("\n")
        self.fe_file.write("bodies \n")
        for k, v in self.cells.items():
            self.fe_file.write(f"{abs(k)}   {k}    VOLUME {self.volume_values[k]} \n")

        self.fe_file.write("\n \n")
        self.fe_file.write("read \n \n")
        self.fe_file.write("show_all_edges off \n")
        # f.write("clipped on \n")
        self.fe_file.write("metric_conversion off \n")
        self.fe_file.write("autorecalc on \n")
        self.fe_file.write("gv_binary off \n")
        self.fe_file.write("gravity off \n")
        self.fe_file.write("ii := 0; \n")

        if not self.polygonal:
            self.add_refining_triangulation(3)
        return self.fe_file

    def add_vertex_averaging(self, how_many: int = 1) -> io.StringIO:
        """
        Add vertex averaging using the V Surface Evolver function, at the end of the Surface Evolver slate

        :param how_many: Number of times the averaging should be done
        :return: Current Surface Evolver slate
        :rtype: io.StringIO()
        """
        self.fe_file.write(f"V {how_many}; \n")
        return self.fe_file

    def add_refining_triangulation(self, how_many: int = 1) -> io.StringIO:
        """
        Add a mesh refinement using the r Surface Evolver function, at the end of the Surface Evolver slate

        :param how_many: Number of times the refinement should be done
        :type how_many: int
        :return: Current Surface Evolver slate
        :rtype: io.StringIO()
        """
        self.fe_file.write(f"r {how_many}; \n")
        return self.fe_file

    def change_scale(self, new_scale: float) -> io.StringIO:
        """
        Change the scale of the Surface Evolver simulation at the end of the Surface Evolver slate

        :param new_scale: Numerical value of the new scale
        :type new_scale: float
        :return: Current Surface Evolver slate
        :rtype: io.StringIO()
        """
        self.fe_file.write(f"scale := {new_scale}; \n")
        return self.fe_file

    def evolve_system(self, steps: int = 1) -> io.StringIO:
        """
        Add evolution function using the Surface Evolver go function

        :param steps: Number of steps to evolve for
        :type steps: int
        :return: Current Surface Evolver slate
        :rtype: io.StringIO()
        """
        self.fe_file.write(f"g {steps}; \n")
        return self.fe_file

    def add_t1_swaps(self, max_size: float = 0.1) -> io.StringIO:
        """
        Add check for T1 swaps using the Surface Evolver t1_edgeswap function

        :param max_size: Maximum size of interfaces before making a T1 swap
        :type max_size: float
        :return: Current Surface Evolver slate
        :rtype: io.StringIO()
        """
        self.fe_file.write(f"t1_edgeswap edge where length < {max_size}; \n")
        return self.fe_file

    def initial_relaxing(self, evolve_step: int = 2500, averaging: int = 100) -> io.StringIO:
        """
        Initial standard relaxing with vertex averaging and scale change followed by evolution

        :param evolve_step: Number of steps to evolve at each scale chnage
        :type evolve_step: int
        :param averaging: Number of vertex averagings to perform
        :type evolve_step: int
        :return: Current Surface Evolver slate
        :rtype: io.StringIO
        """
        self.add_vertex_averaging(averaging)
        self.change_scale(0.25)
        self.evolve_system(evolve_step)
        self.add_vertex_averaging(averaging)
        self.change_scale(0.1)
        self.evolve_system(evolve_step)
        self.add_vertex_averaging(averaging)
        self.change_scale(0.01)
        return self.fe_file

    def evolve_relaxing(self, number_of_times: int = 1, steps: int = 1, max_size: float = 0.1) -> io.StringIO:
        """
        Evolve the system a fixed number of steps and perform T1 swaps after a definite number of times

        :param number_of_times: Number of times the evolution plus T1 swaps happen
        :type number_of_times: int
        :param steps: Number of evolution step for all iterations
        :type steps: int
        :param max_size: Maximum size allowed for membranes before a T1 happens
        :type max_size: float
        :return: Current Surface Evolver slate
        :rtype: io.StringIO
        """
        for _ in range(0, number_of_times):
            self.evolve_system(steps)
            self.add_t1_swaps(max_size)
        return self.fe_file

    def save_one_step(self, output_directory: str, file_name: str) -> io.StringIO:
        """
        Add a savepoint to the Surface Evolver slate using the sprintf function

        :param output_directory: Folder to save the file into
        :type output_directory: str
        :param file_name: Name of the file to be saved
        :type file_name: str
        :return: Current Surface Evolver slate
        :rtype: io.StringIO
        """
        self.fe_file.write(f'ff := sprintf "{output_directory}/{file_name}%d.dmp",ii; dump ff; ii+=1; \n')
        return self.fe_file

    def save_many_steps(self, output_directory, file_name, max_steps, time_step=1, averaging=1, max_size=0.1):
        self.fe_file.write('while ii < ' + str(max_steps) + ' do { '
                                                            'g ' + str(time_step) + '; V ' + str(averaging) +
                           '; t1_edgeswap edge where length < 0.1; '
                           'ff := sprintf "' + output_directory + "/" + file_name + '%d.dmp",ii;'
                                                                                    ' dump ff; ii:=ii+1} \n')
        return self.fe_file

    def save_fe_file(self, file_name: str) -> bool:
        """
        Save the Surface Evolver slate to disk

        :param file_name:
        :return: Success state of the saving
        :rtype: bool
        """
        self.fe_file.write('q; \n')
        with open(f'{file_name}', mode='w') as f:
            print(self.fe_file.getvalue(), file=f)
        return True

    def change_line_tensions(self, new_tensions: dict) -> io.StringIO:
        """
        Set new densities for the system's  membranes. Setting is done using the dictionary key as membrane id and the
        value as the new tension

        :param new_tensions:
        :return: Current Surface Evolver slate
        :rtype: io.StringIO
        """
        for eid, tension in new_tensions.items():
            self.fe_file.write(f"set edges density {tension} where original == {abs(eid)}; \n")
        return self.fe_file
